<?php /** @noinspection ALL */

namespace Slovocast\Domain\Repository\Channel;

use DateTime;
use DateTimeImmutable;
use Exception;
use Slovocast\Domain\Repository\ChannelRepositoryInterface;
use Slovocast\Domain\Entity\Channel;
use Slovocast\Domain\Entity\User;
use Slovocast\Exception\EntityNotFoundException;
use React\Mysql\MysqlClient;
use Throwable;
use function React\Async\await;

class ChannelRepository implements ChannelRepositoryInterface
{
    public function __construct(
        protected MysqlClient $db
    ) { }

    /**
     * @param array $results Result Set from the Query
     * @return Channel
     * @throws Exception
     */
    protected function createChannelFromResults(array $results): Channel
    {
        return Channel::fromArray([
            'id' => $results['id'],
            'name' => $results['name'],
            'slug' => $results['slug'],
            'description' => $results['description'],
            'link' => $results['link'],
            'language' => $results['language'],
            'copyright' => $results['copyright'],
            'explicit' => $results['explicit'],
            'createdAt' => new DateTimeImmutable($results['created_at']),
            'updatedAt' => new DateTime($results['updated_at'])
        ]);
    }

    /**
     * @param int $id
     * @return Channel
     * @throws EntityNotFoundException
     * @throws Throwable
     */
    public function get(int $id): Channel
    {
        $query = "SELECT * FROM channels WHERE id = ?";
        $results = await($this->db->query($query, [ $id ]));

        if (!count($results->resultRows)) {
            throw new EntityNotFoundException("Unable to find Channel");
        }

        return $this->createChannelFromResults($results->resultRows[0]);
    }

    /**
     * @throws EntityNotFoundException
     * @throws Throwable
     */
    public function getFromUser(User $user): Channel
    {
        $query = "SELECT * FROM channels WHERE owner_id = ?";
        $results = await($this->db->query($query, [ $user->getId() ]));

        if (!count($results->resultRows)) {
            throw new EntityNotFoundException("Unable to find Channel");
        }

        return $this->createChannelFromResults($results->resultRows[0]);
    }

    /**
     * @throws EntityNotFoundException
     * @throws Throwable
     */
    public function getFromSlug(string $slug): Channel
    {
        $query = "SELECT * FROM channels WHERE slug = ?";
        $results = await($this->db->query($query, [ $slug ]));

        if (!count($results->resultRows)) {
            throw new EntityNotFoundException("Unable to find Channel");
        }

        return $this->createChannelFromResults($results->resultRows[0]);
    }

    /**
     * @throws Throwable
     */
    public function create(Channel $channel, User $owner): int
    {
        $query = <<<SQL
INSERT INTO channels (name, slug, description, link,
                      language, copyright, explicit, owner_id)
VALUES (?, ?, ?, ?, ?, ?, ?, ?)
SQL;

        $results = await($this->db->query($query, [
            $channel->getName(),
            $channel->getSlug(),
            $channel->getDescription(),
            $channel->getLink(),
            $channel->getLanguage(),
            $channel->getCopyright(),
            $channel->isExplicit(),
            $owner->getId()
        ]));

        return $results->insertId;
    }

    /**
     * @param Channel $channel
     * @return bool
     * @throws Throwable
     */
    public function update(Channel $channel): bool
    {
        $query = <<<SQL
UPDATE channels SET name = ?,
                    slug = ?,
                    description = ?,
                    link = ?,
                    language = ?,
                    copyright = ?,
                    explicit = ?
WHERE id = ?;
SQL;

        $results = await($this->db->query($query, [
            $channel->getName(),
            $channel->getSlug(),
            $channel->getDescription(),
            $channel->getLink(),
            $channel->getLanguage(),
            $channel->getCopyright(),
            $channel->isExplicit(),
            $channel->getId(),
        ]));

        return (bool) $results->affectedRows;
    }

    public function delete(Channel $channel): bool
    {
        return false;
    }
}
